@@ -8,15 +8,17 @@ use std::fs::File;
 use std::io::Read;
 use std::path::{Path as FilePath, PathBuf as FilePathBuf};
 
+use syn::ext::IdentExt;
+
 use crate::bindgen::bitflags;
 use crate::bindgen::cargo::{Cargo, PackageRef};
 use crate::bindgen::config::{Config, ParseConfig};
 use crate::bindgen::error::Error;
 use crate::bindgen::ir::{
-    AnnotationSet, Cfg, Constant, Documentation, Enum, Function, GenericParams, ItemMap,
-    OpaqueItem, Path, Static, Struct, Type, Typedef, Union,
+    AnnotationSet, AnnotationValue, Cfg, Constant, Documentation, Enum, Function, GenericParam,
+    GenericParams, ItemMap, OpaqueItem, Path, Static, Struct, Type, Typedef, Union,
 };
-use crate::bindgen::utilities::{SynAbiHelpers, SynAttributeHelpers, SynItemFnHelpers};
+use crate::bindgen::utilities::{SynAbiHelpers, SynAttributeHelpers, SynItemHelpers};
 
 const STD_CRATES: &[&str] = &[
     "std",
@@ -55,6 +57,7 @@ pub fn parse_src(src_file: &FilePath, config: &Config) -> ParseResult {
     };
 
     context.parse_mod(&pkg_ref, src_file, 0)?;
+    context.out.source_files = context.cache_src.keys().map(|k| k.to_owned()).collect();
     Ok(context.out)
 }
 
@@ -77,6 +80,14 @@ pub(crate) fn parse_lib(lib: Cargo, config: &Config) -> ParseResult {
 
     let binding_crate = context.lib.as_ref().unwrap().binding_crate_ref();
     context.parse_crate(&binding_crate)?;
+    context.out.source_files = context.cache_src.keys().map(|k| k.to_owned()).collect();
+    context.out.package_version = context
+        .lib
+        .as_ref()
+        .unwrap()
+        .binding_crate_ref()
+        .version
+        .unwrap();
     Ok(context.out)
 }
 
@@ -95,7 +106,7 @@ struct Parser<'a> {
     out: Parse,
 }
 
-impl<'a> Parser<'a> {
+impl Parser<'_> {
     fn should_parse_dependency(&self, pkg_name: &str) -> bool {
         if self.parsed_crates.contains(pkg_name) {
             return false;
@@ -154,14 +165,16 @@ impl<'a> Parser<'a> {
                 None => {
                     // This should be an error, but is common enough to just elicit a warning
                     warn!(
-                        "Parsing crate `{}`: can't find lib.rs with `cargo metadata`.",
+                        "Parsing crate `{}`: can't find lib.rs with `cargo metadata`. \
+                        The crate may be available only on a particular platform, \
+                        so consider setting `fetch_all_dependencies` in your cbindgen configuration.",
                         pkg.name
                     );
                 }
             }
         }
 
-        for (dep_pkg, cfg) in self.lib.as_ref().unwrap().dependencies(&pkg) {
+        for (dep_pkg, cfg) in self.lib.as_ref().unwrap().dependencies(pkg) {
             if !self.should_parse_dependency(&dep_pkg.name) {
                 continue;
             }
@@ -183,13 +196,6 @@ impl<'a> Parser<'a> {
     fn parse_expand_crate(&mut self, pkg: &PackageRef) -> Result<(), Error> {
         assert!(self.lib.is_some());
 
-        // If you want to expand the crate you run cbindgen on you might end up in an endless
-        // recursion if the cbindgen generation is triggered from build.rs. Hence don't run the
-        // expansion if the build was already triggered by cbindgen.
-        if std::env::var("_CBINDGEN_IS_RUNNING").is_ok() {
-            return Ok(());
-        }
-
         let mod_items = {
             if !self.cache_expanded_crate.contains_key(&pkg.name) {
                 let s = self
@@ -298,7 +304,7 @@ impl<'a> Parser<'a> {
         debug_assert_eq!(mod_dir.is_some(), submod_dir.is_some());
         // We process the items first then the nested modules.
         let nested_modules = self.out.load_syn_crate_mod(
-            &self.config,
+            self.config,
             &self.binding_crate_name,
             &pkg.name,
             Cfg::join(&self.cfg_stack).as_ref(),
@@ -306,7 +312,7 @@ impl<'a> Parser<'a> {
         );
 
         for item in nested_modules {
-            let next_mod_name = item.ident.to_string();
+            let next_mod_name = item.ident.unraw().to_string();
             let cfg = Cfg::load(&item.attrs);
             if let Some(ref cfg) = cfg {
                 self.cfg_stack.push(cfg.clone());
@@ -339,8 +345,11 @@ impl<'a> Parser<'a> {
                     // Last chance to find a module path
                     let mut path_attr_found = false;
                     for attr in &item.attrs {
-                        if let Ok(syn::Meta::NameValue(syn::MetaNameValue { path, lit, .. })) =
-                            attr.parse_meta()
+                        if let syn::Meta::NameValue(syn::MetaNameValue {
+                            path,
+                            value: syn::Expr::Lit(syn::ExprLit { lit, .. }),
+                            ..
+                        }) = &attr.meta
                         {
                             match lit {
                                 syn::Lit::Str(ref path_lit) if path.is_ident("path") => {
@@ -409,6 +418,8 @@ pub struct Parse {
     pub opaque_items: ItemMap<OpaqueItem>,
     pub typedefs: ItemMap<Typedef>,
     pub functions: Vec<Function>,
+    pub source_files: Vec<FilePathBuf>,
+    pub package_version: String,
 }
 
 impl Parse {
@@ -422,13 +433,18 @@ impl Parse {
             opaque_items: ItemMap::default(),
             typedefs: ItemMap::default(),
             functions: Vec::new(),
+            source_files: Vec::new(),
+            package_version: String::new(),
         }
     }
 
     pub fn add_std_types(&mut self) {
         let mut add_opaque = |path: &str, generic_params: Vec<&str>| {
             let path = Path::new(path);
-            let generic_params: Vec<_> = generic_params.into_iter().map(Path::new).collect();
+            let generic_params: Vec<_> = generic_params
+                .into_iter()
+                .map(GenericParam::new_type_param)
+                .collect();
             self.opaque_items.try_insert(OpaqueItem::new(
                 path,
                 GenericParams(generic_params),
@@ -466,6 +482,8 @@ impl Parse {
         self.opaque_items.extend_with(&other.opaque_items);
         self.typedefs.extend_with(&other.typedefs);
         self.functions.extend_from_slice(&other.functions);
+        self.source_files.extend_from_slice(&other.source_files);
+        self.package_version.clone_from(&other.package_version);
     }
 
     fn load_syn_crate_mod<'a>(
@@ -515,10 +533,10 @@ impl Parse {
                     self.load_syn_ty(crate_name, mod_cfg, item);
                 }
                 syn::Item::Impl(ref item_impl) => {
-                    let has_assoc_const = item_impl.items.iter().any(|item| match item {
-                        syn::ImplItem::Const(_) => true,
-                        _ => false,
-                    });
+                    let has_assoc_const = item_impl
+                        .items
+                        .iter()
+                        .any(|item| matches!(item, syn::ImplItem::Const(_)));
                     if has_assoc_const {
                         impls_with_assoc_consts.push(item_impl);
                     }
@@ -526,7 +544,9 @@ impl Parse {
                     if let syn::Type::Path(ref path) = *item_impl.self_ty {
                         if let Some(type_name) = path.path.get_ident() {
                             for method in item_impl.items.iter().filter_map(|item| match item {
-                                syn::ImplItem::Method(method) => Some(method),
+                                syn::ImplItem::Fn(method) if !method.should_skip_parsing() => {
+                                    Some(method)
+                                }
                                 _ => None,
                             }) {
                                 self.load_syn_method(
@@ -534,7 +554,7 @@ impl Parse {
                                     binding_crate_name,
                                     crate_name,
                                     mod_cfg,
-                                    &Path::new(type_name.to_string()),
+                                    &Path::new(type_name.unraw().to_string()),
                                     method,
                                 )
                             }
@@ -542,7 +562,7 @@ impl Parse {
                     }
                 }
                 syn::Item::Macro(ref item) => {
-                    self.load_builtin_macro(config, crate_name, mod_cfg, item)
+                    self.load_builtin_macro(config, crate_name, mod_cfg, item);
                 }
                 syn::Item::Mod(ref item) => {
                     nested_modules.push(item);
@@ -565,7 +585,11 @@ impl Parse {
         item_impl: &syn::ItemImpl,
     ) {
         let associated_constants = item_impl.items.iter().filter_map(|item| match item {
-            syn::ImplItem::Const(ref associated_constant) => Some(associated_constant),
+            syn::ImplItem::Const(ref associated_constant)
+                if !associated_constant.should_skip_parsing() =>
+            {
+                Some(associated_constant)
+            }
             _ => None,
         });
         self.load_syn_assoc_consts(
@@ -585,11 +609,12 @@ impl Parse {
         mod_cfg: Option<&Cfg>,
         item: &syn::ItemForeignMod,
     ) {
-        if !item.abi.is_c() {
+        if !item.abi.is_c() && !item.abi.is_omitted() {
             info!("Skip {} - (extern block must be extern C).", crate_name);
             return;
         }
 
+        let mod_cfg = Cfg::append(mod_cfg, Cfg::load(&item.attrs));
         for foreign_item in &item.items {
             if let syn::ForeignItem::Fn(ref function) = *foreign_item {
                 if !config
@@ -602,8 +627,15 @@ impl Parse {
                     );
                     return;
                 }
-                let path = Path::new(function.sig.ident.to_string());
-                match Function::load(path, None, &function.sig, true, &function.attrs, mod_cfg) {
+                let path = Path::new(function.sig.ident.unraw().to_string());
+                match Function::load(
+                    path,
+                    None,
+                    &function.sig,
+                    true,
+                    &function.attrs,
+                    mod_cfg.as_ref(),
+                ) {
                     Ok(func) => {
                         info!("Take {}::{}.", crate_name, &function.sig.ident);
 
@@ -628,7 +660,7 @@ impl Parse {
         crate_name: &str,
         mod_cfg: Option<&Cfg>,
         self_type: &Path,
-        item: &syn::ImplItemMethod,
+        item: &syn::ImplItemFn,
     ) {
         self.load_fn_declaration(
             config,
@@ -638,7 +670,6 @@ impl Parse {
             item,
             Some(self_type),
             &item.sig,
-            &item.vis,
             &item.attrs,
         )
     }
@@ -660,7 +691,6 @@ impl Parse {
             item,
             None,
             &item.sig,
-            &item.vis,
             &item.attrs,
         );
     }
@@ -672,10 +702,9 @@ impl Parse {
         binding_crate_name: &str,
         crate_name: &str,
         mod_cfg: Option<&Cfg>,
-        named_symbol: &dyn SynItemFnHelpers,
+        named_symbol: &dyn SynItemHelpers,
         self_type: Option<&Path>,
         sig: &syn::Signature,
-        vis: &syn::Visibility,
         attrs: &[syn::Attribute],
     ) {
         if !config
@@ -695,60 +724,36 @@ impl Parse {
             if let Some(ref self_type) = self_type {
                 items.push(self_type.to_string());
             }
-            items.push(sig.ident.to_string());
+            items.push(sig.ident.unraw().to_string());
             items.join("::")
         };
 
         let is_extern_c = sig.abi.is_omitted() || sig.abi.is_c();
         let exported_name = named_symbol.exported_name();
 
-        if let syn::Visibility::Public(_) = vis {
-            match (is_extern_c, exported_name) {
-                (true, Some(exported_name)) => {
-                    let path = Path::new(exported_name);
-                    match Function::load(path, self_type, &sig, false, &attrs, mod_cfg) {
-                        Ok(func) => {
-                            info!("Take {}.", loggable_item_name());
-                            self.functions.push(func);
-                        }
-                        Err(msg) => {
-                            error!("Cannot use fn {} ({}).", loggable_item_name(), msg);
-                        }
+        match (is_extern_c, exported_name) {
+            (true, Some(exported_name)) => {
+                let path = Path::new(exported_name);
+                match Function::load(path, self_type, sig, false, attrs, mod_cfg) {
+                    Ok(func) => {
+                        info!("Take {}.", loggable_item_name());
+                        self.functions.push(func);
+                    }
+                    Err(msg) => {
+                        error!("Cannot use fn {} ({}).", loggable_item_name(), msg);
                     }
                 }
-                (true, None) => {
-                    warn!(
-                        "Skipping {} - (not `no_mangle`, and has no `export_name` attribute)",
-                        loggable_item_name()
-                    );
-                }
-                (false, Some(_exported_name)) => {
-                    warn!("Skipping {} - (not `extern \"C\"`", loggable_item_name());
-                }
-                (false, None) => {}
             }
-        } else {
-            match (is_extern_c, exported_name) {
-                (true, Some(..)) => {
-                    warn!(
-                        "Skipping {} - (not `pub` but is `extern \"C\"` and `no_mangle`)",
-                        loggable_item_name()
-                    );
-                }
-                (true, None) => {
-                    warn!(
-                        "Skipping {} - (not `pub` but is `extern \"C\"`)",
-                        loggable_item_name()
-                    );
-                }
-                (false, Some(..)) => {
-                    warn!(
-                        "Skipping {} - (not `pub` but is `no_mangle`)",
-                        loggable_item_name()
-                    );
-                }
-                (false, None) => {}
+            (true, None) => {
+                warn!(
+                    "Skipping {} - (not `no_mangle`, and has no `export_name` attribute)",
+                    loggable_item_name()
+                );
             }
+            (false, Some(_exported_name)) => {
+                warn!("Skipping {} - (not `extern \"C\"`)", loggable_item_name());
+            }
+            (false, None) => {}
         }
     }
 
@@ -793,7 +798,7 @@ impl Parse {
                 return;
             }
 
-            let path = Path::new(item.ident.to_string());
+            let path = Path::new(item.ident.unraw().to_string());
             match Constant::load(
                 path,
                 mod_cfg,
@@ -851,7 +856,7 @@ impl Parse {
             return;
         }
 
-        let path = Path::new(item.ident.to_string());
+        let path = Path::new(item.ident.unraw().to_string());
         match Constant::load(path, mod_cfg, &item.ty, &item.expr, &item.attrs, None) {
             Ok(constant) => {
                 info!("Take {}::{}.", crate_name, &item.ident);
@@ -887,27 +892,18 @@ impl Parse {
             return;
         }
 
-        if let syn::Visibility::Public(_) = item.vis {
-            if item.is_no_mangle() {
-                match Static::load(item, mod_cfg) {
-                    Ok(constant) => {
-                        info!("Take {}::{}.", crate_name, &item.ident);
-
-                        self.globals.try_insert(constant);
-                    }
-                    Err(msg) => {
-                        warn!("Skip {}::{} - ({})", crate_name, &item.ident, msg);
-                    }
+        if let Some(exported_name) = item.exported_name() {
+            let path = Path::new(exported_name);
+            match Static::load(path, item, mod_cfg) {
+                Ok(constant) => {
+                    info!("Take {}::{}.", crate_name, &item.ident);
+                    self.globals.try_insert(constant);
+                }
+                Err(msg) => {
+                    warn!("Skip {}::{} - ({})", crate_name, &item.ident, msg);
                 }
             }
-        }
-
-        // TODO
-        if let syn::Visibility::Public(_) = item.vis {
         } else {
-            warn!("Skip {}::{} - (not `pub`).", crate_name, &item.ident);
-        }
-        if !item.is_no_mangle() {
             warn!("Skip {}::{} - (not `no_mangle`).", crate_name, &item.ident);
         }
     }
@@ -927,7 +923,7 @@ impl Parse {
             }
             Err(msg) => {
                 info!("Take {}::{} - opaque ({}).", crate_name, &item.ident, msg);
-                let path = Path::new(item.ident.to_string());
+                let path = Path::new(item.ident.unraw().to_string());
                 self.opaque_items.try_insert(
                     OpaqueItem::load(path, &item.generics, &item.attrs, mod_cfg).unwrap(),
                 );
@@ -951,7 +947,7 @@ impl Parse {
             }
             Err(msg) => {
                 info!("Take {}::{} - opaque ({}).", crate_name, &item.ident, msg);
-                let path = Path::new(item.ident.to_string());
+                let path = Path::new(item.ident.unraw().to_string());
                 self.opaque_items.try_insert(
                     OpaqueItem::load(path, &item.generics, &item.attrs, mod_cfg).unwrap(),
                 );
@@ -974,7 +970,7 @@ impl Parse {
             }
             Err(msg) => {
                 info!("Take {}::{} - opaque ({}).", crate_name, &item.ident, msg);
-                let path = Path::new(item.ident.to_string());
+                let path = Path::new(item.ident.unraw().to_string());
                 self.opaque_items.try_insert(
                     OpaqueItem::load(path, &item.generics, &item.attrs, mod_cfg).unwrap(),
                 );
@@ -992,7 +988,7 @@ impl Parse {
             }
             Err(msg) => {
                 info!("Take {}::{} - opaque ({}).", crate_name, &item.ident, msg);
-                let path = Path::new(item.ident.to_string());
+                let path = Path::new(item.ident.unraw().to_string());
                 self.opaque_items.try_insert(
                     OpaqueItem::load(path, &item.generics, &item.attrs, mod_cfg).unwrap(),
                 );
@@ -1008,7 +1004,7 @@ impl Parse {
         item: &syn::ItemMacro,
     ) {
         let name = match item.mac.path.segments.last() {
-            Some(ref n) => n.ident.to_string(),
+            Some(n) => n.ident.unraw().to_string(),
             None => return,
         };
 
@@ -1017,7 +1013,7 @@ impl Parse {
         }
 
         let bitflags = match bitflags::parse(item.mac.tokens.clone()) {
-            Ok(b) => b,
+            Ok(bf) => bf,
             Err(e) => {
                 warn!("Failed to parse bitflags invocation: {:?}", e);
                 return;
@@ -1025,10 +1021,18 @@ impl Parse {
         };
 
         let (struct_, impl_) = bitflags.expand();
-        self.load_syn_struct(config, crate_name, mod_cfg, &struct_);
-        // We know that the expansion will only reference `struct_`, so it's
-        // fine to just do it here instead of deferring it like we do with the
-        // other calls to this function.
-        self.load_syn_assoc_consts_from_impl(crate_name, mod_cfg, &impl_);
+        if let Some(struct_) = struct_ {
+            self.load_syn_struct(config, crate_name, mod_cfg, &struct_);
+        }
+        if let syn::Type::Path(ref path) = *impl_.self_ty {
+            if let Some(type_name) = path.path.get_ident() {
+                self.structs
+                    .for_items_mut(&Path::new(type_name.unraw().to_string()), |item| {
+                        item.annotations
+                            .add_default("internal-derive-bitflags", AnnotationValue::Bool(true));
+                    });
+            }
+        }
+        self.load_syn_assoc_consts_from_impl(crate_name, mod_cfg, &impl_)
     }
 }
