@@ -6,9 +6,9 @@
 
 namespace constants {
 
-static const int32_t FOO = 10;
+constexpr static const int32_t FOO = 10;
 
-static const float ZOM = 3.14;
+constexpr static const float ZOM = 3.14;
 
 struct Foo {
   int32_t x[FOO];
@@ -18,6 +18,6 @@ extern "C" {
 
 void root(Foo x);
 
-} // extern "C"
+}  // extern "C"
 
-} // namespace constants
+}  // namespace constants
